import os.path
import re
import sys
import tokenize
import typing

import click
import yaml

from ecs_deplojo.connection import Connection
from ecs_deplojo.deployment import DeploymentFailed, start_deployment
from ecs_deplojo.logger import logger
from ecs_deplojo.task_definitions import generate_task_definitions


class VarType(click.ParamType):
    name = "var"
    re_pattern = re.compile("^%s$" % tokenize.Name)

    def convert(self, value, param, ctx):
        try:
            key, value = value.split("=", 1)
            if not self.re_pattern.match(key):
                self.fail("%s is not a valid identifier" % key)
        except ValueError:
            self.fail("%s is not a valid key/value string" % value, param, ctx)

        return (key, value)


@click.command()
@click.option("--config", required=True, type=click.Path())
@click.option("--var", multiple=True, type=VarType())
@click.option("--dry-run", is_flag=True, default=False)
@click.option("--output-path", required=False, type=click.Path())
@click.option("--role-arn", required=False, type=str)
@click.option("--create-missing-services", default=False, type=bool)
def main(
    config, var, output_path, dry_run, role_arn=None, create_missing_services=False
):
    try:
        run(
            filename=config,
            template_vars=dict(var),
            role_arn=role_arn,
            output_path=output_path,
            create_missing_services=create_missing_services,
            dry_run=dry_run,
        )
    except DeploymentFailed:
        sys.exit(1)


def run(
    filename: str,
    template_vars: typing.Dict[str, str],
    role_arn: typing.Optional[str] = None,
    output_path: typing.Optional[str] = None,
    create_missing_services=False,
    dry_run=False,
):
    base_path = os.path.dirname(filename)
    with open(filename, "r") as fh:
        config = yaml.safe_load(fh.read())

    connection = Connection(role_arn)
    cluster_name = config["cluster_name"]
    services = config["services"]
    logger.info(
        "Starting deploy on cluster %s (%s services)", cluster_name, len(services)
    )

    # Generate the task definitions
    task_definitions = generate_task_definitions(
        config, template_vars, base_path, output_path
    )

    # Check if all task definitions required by the services exists
    for service_name, service in services.items():
        if service["task_definition"] not in task_definitions:
            logger.error(
                "Missing task definition %r for service %r",
                service["task_definition"],
                service_name,
            )

    # Run the deployment
    if not dry_run:
        try:
            start_deployment(
                config, connection, task_definitions, create_missing_services
            )
        except DeploymentFailed:
            logger.exception("Error, exiting")
            sys.exit(1)



import json
import os
from textwrap import dedent

import boto3
import moto
import pytest
from moto.ec2 import utils as ec2_utils

from ecs_deplojo.connection import Connection
from ecs_deplojo.task_definitions import TaskDefinition

BASE_DIR = os.path.dirname(os.path.abspath(__file__))


@pytest.yield_fixture(scope="function")
def cluster():
    with moto.mock_aws():
        boto3.setup_default_session(region_name="eu-west-1")

        ec2 = boto3.client("ec2", region_name="eu-west-1")
        ecs = boto3.client("ecs", region_name="eu-west-1")

        known_amis = ec2.describe_images()
        image_id = known_amis['Images'][0]['ImageId']

        ec2 = boto3.resource("ec2", region_name="eu-west-1")

        test_instance = ec2.create_instances(
            ImageId=image_id, MinCount=1, MaxCount=1
        )[0]

        instance_id_document = json.dumps(
            ec2_utils.generate_instance_identity_document(test_instance)
        )

        cluster = ecs.create_cluster(clusterName="default")
        ecs.register_container_instance(
            cluster="default", instanceIdentityDocument=instance_id_document
        )

        yield cluster


@pytest.fixture
def connection(cluster):
    return Connection()


@pytest.fixture
def definition():
    path = os.path.join(BASE_DIR, "files/default_taskdef.json")

    with open(path, "r") as json_file:
        return TaskDefinition(json.load(json_file))


@pytest.fixture
def default_config():
    path = os.path.join(BASE_DIR, "files/default_config.yml")

    with open(path, "r") as fh:
        yield fh


@pytest.fixture
def example_project(tmpdir):
    data = """
    {
      "family": "default",
      "volumes": [],
      "containerDefinitions": [
        {
          "name": "web-1",
          "image": "${image}",
          "essential": true,
          "command": ["hello", "world"],
          "memory": 256,
          "cpu": 0,
          "portMappings": [
            {
              "containerPort": 8080,
              "hostPort": 0
            }
          ]
        },
        {
          "name": "web-2",
          "image": "${image}",
          "essential": true,
          "command": ["hello", "world"],
          "memory": 256,
          "cpu": 0,
          "portMappings": [
            {
              "containerPort": 8080,
              "hostPort": 0
            }
          ]
        }
      ]
    }
    """.strip()

    filename = tmpdir.join("task_definition.json")
    filename.write(data)

    data = dedent(
        """
    ---
    cluster_name: default
    environment:
      DATABASE_URL: postgresql://
    environment_groups:
      group-1:
        ENV_CODE: 12345
    task_definitions:
      web:
        template: %(template_filename)s
        environment_group: group-1
        task_role_arn: my-test
        overrides:
          web-1:
            memory: 512
            portMappings:
              - hostPort: 0
                containerPort: 8080
                protocol: tcp
    services:
      web:
        task_definition: web

    before_deploy:
      - task_definition: web
        container: web-1
        command: manage.py migrate --noinput

    after_deploy:
      - task_definition: web
        container: web-1
        command: manage.py clearsessions
    """
        % {"template_filename": filename.strpath}
    )

    filename = tmpdir.join("config.yml")
    filename.write(data)
    return filename



import pytest
from click.testing import CliRunner

from ecs_deplojo import cli


def test_cli_execution_existing_service(
    example_project, cluster, caplog, connection, definition
):
    # Make sure the service exists
    retval = connection.ecs.register_task_definition(**definition.as_dict())
    task_definition_arn = retval["taskDefinition"]["taskDefinitionArn"]
    connection.ecs.create_service(
        cluster=cluster["cluster"]["clusterName"],
        serviceName="web",
        taskDefinition=task_definition_arn,
        desiredCount=1,
    )

    runner = CliRunner()
    result = runner.invoke(
        cli.main,
        ["--config=%s" % example_project.strpath, "--var=image=my-docker-image:1.0"],
    )
    assert result.exit_code == 0, result.output

    expected = [
        "Starting deploy on cluster default (1 services)",
        "Registered new task definition web:1",
        "Starting one-off task 'manage.py migrate --noinput' via web:1 (web-1)",
        "Updating service web with task definition web:1",
        "Waiting for deployments",
        "Deployment finished: web (1/1)",
        "Starting one-off task 'manage.py clearsessions' via web:1 (web-1)",
        "Deregistering old task definitions",
        " - web",
    ]
    lines = [r.message for r in caplog.records if r.name.startswith("deploy")]
    assert lines == expected


def test_run_missing_service(example_project, cluster, caplog):
    with pytest.raises(SystemExit):
        cli.run(
            filename=example_project.strpath,
            template_vars={"image": "my-docker-image:1.0"},
            create_missing_services=False,
        )


def test_run_create_service(example_project, cluster, caplog):
    cli.run(
        filename=example_project.strpath,
        template_vars={"image": "my-docker-image:1.0"},
        create_missing_services=True,
    )

    expected = [
        "Starting deploy on cluster default (1 services)",
        "Registered new task definition web:1",
        "Starting one-off task 'manage.py migrate --noinput' via web:1 (web-1)",
        "Creating new service web with task definition web:1",
        "Waiting for deployments",
        "Deployment finished: web (1/1)",
        "Starting one-off task 'manage.py clearsessions' via web:1 (web-1)",
        "Deregistering old task definitions",
        " - web",
    ]
    lines = [r.message for r in caplog.records if r.name.startswith("deploy")]
    assert lines == expected


def test_run_update_service(example_project, cluster, connection, definition, caplog):
    # Make sure the service exists
    retval = connection.ecs.register_task_definition(**definition.as_dict())
    task_definition_arn = retval["taskDefinition"]["taskDefinitionArn"]
    connection.ecs.create_service(
        cluster=cluster["cluster"]["clusterName"],
        serviceName="web",
        taskDefinition=task_definition_arn,
        desiredCount=1,
    )
    cli.run(
        filename=example_project.strpath,
        template_vars={"image": "my-docker-image:1.0"},
        create_missing_services=False,
    )

    expected = [
        "Starting deploy on cluster default (1 services)",
        "Registered new task definition web:1",
        "Starting one-off task 'manage.py migrate --noinput' via web:1 (web-1)",
        "Updating service web with task definition web:1",
        "Waiting for deployments",
        "Deployment finished: web (1/1)",
        "Starting one-off task 'manage.py clearsessions' via web:1 (web-1)",
        "Deregistering old task definitions",
        " - web",
    ]
    lines = [r.message for r in caplog.records if r.name.startswith("deploy")]
    assert lines == expected



import copy
import os

from ecs_deplojo import register
from ecs_deplojo.task_definitions import TaskDefinition

BASE_DIR = os.path.dirname(os.path.abspath(__file__))


def test_transform_definition(definition):
    definition.container_definitions[0]["environment"] = {
        "DEBUG": True,
        "AWS_REGION": "eu-west-1",
    }
    result = definition.as_dict()

    assert result["containerDefinitions"][0]["environment"] == [
        {"name": "AWS_REGION", "value": "eu-west-1"},
        {"name": "DEBUG", "value": "True"},
    ]


def test_register_task_definitions(cluster, connection):
    task_definitions = {
        "service-1": TaskDefinition(
            {
                "family": "my-task-def",
                "volumes": [],
                "containerDefinitions": [
                    {
                        "name": "default",
                        "image": "my-docker-image:1.0",
                        "essential": True,
                        "command": ["hello", "world"],
                        "hostname": "my-task-def",
                        "memory": 256,
                        "cpu": 0,
                        "portMappings": [{"containerPort": 8080, "hostPort": 0}],
                        "environment": {},
                    }
                ],
                "tags": [{"key": "createdBy", "value": "ecs-deplojo"}],
            }
        )
    }

    result = connection.ecs.list_task_definitions()
    assert len(result["taskDefinitionArns"]) == 0

    register.register_task_definitions(connection, task_definitions)

    result = connection.ecs.list_task_definitions()
    assert len(result["taskDefinitionArns"]) == 1


def test_deregister_task_definitions(cluster, connection):
    task_definitions = {
        "service-1": TaskDefinition(
            {
                "family": "my-task-def",
                "volumes": [],
                "containerDefinitions": [
                    {
                        "name": "default",
                        "image": "my-docker-image:1.0",
                        "essential": True,
                        "command": ["hello", "world"],
                        "hostname": "my-task-def",
                        "memory": 256,
                        "cpu": 0,
                        "portMappings": [{"containerPort": 8080, "hostPort": 0}],
                        "environment": {},
                    }
                ],
                "tags": [{"key": "createdBy", "value": "ecs-deplojo"}],
            }
        )
    }

    result = connection.ecs.list_task_definitions()
    assert len(result["taskDefinitionArns"]) == 0

    for i in range(10):
        task_def = copy.deepcopy(task_definitions)
        register.register_task_definitions(connection, task_def)

    result = connection.ecs.list_task_definitions()
    assert len(result["taskDefinitionArns"]) == 10

    register.deregister_task_definitions(connection, task_def)
    result = connection.ecs.list_task_definitions()
    # deregistration of task definitions doesn't appear to work
    #assert len(result["taskDefinitionArns"]) == 1
    assert len(result["taskDefinitionArns"]) > 0



import typing

from ecs_deplojo.connection import Connection
from ecs_deplojo.logger import logger
from ecs_deplojo.task_definitions import TaskDefinition


def register_task_definitions(
    connection: Connection, task_definitions: typing.Dict[str, TaskDefinition]
) -> None:
    """Update task definitions"""

    for service_name, task_definition in task_definitions.items():
        definition = task_definition.as_dict()
        result = connection.ecs.register_task_definition(**definition)

        task_definition.family = result["taskDefinition"]["family"]
        task_definition.revision = result["taskDefinition"]["revision"]
        task_definition.name = "%s:%s" % (
            result["taskDefinition"]["family"],
            result["taskDefinition"]["revision"],
        )
        task_definition.arn = result["taskDefinition"]["taskDefinitionArn"]
        logger.info("Registered new task definition %s", task_definition)


def update_scheduled_tasks(
    connection: Connection,
    task_definitions: dict[str, TaskDefinition],
    scheduled_tasks: dict[str, typing.Any],
):
    for rule_name, config in scheduled_tasks.items():
        if not (task_definition := config.get("task_definition")):
            continue
        if not (task_definition := task_definitions.get(task_definition)):
            continue

        targets = connection.events.list_targets_by_rule(Rule=rule_name)
        for target in targets["Targets"]:
            target["EcsParameters"]["TaskDefinitionArn"] = task_definition.arn

        connection.events.put_targets(Rule=rule_name, Targets=targets["Targets"])


def deregister_task_definitions(
    connection: Connection, task_definitions: typing.Dict[str, TaskDefinition]
) -> None:
    """Deregister all task definitions not used currently which are created
    by ecs-deplojo.

    """

    def yield_arns(family) -> typing.Generator[str, None, None]:
        paginator = connection.ecs.get_paginator("list_task_definitions")
        for page in paginator.paginate(familyPrefix=family):
            for arn in page["taskDefinitionArns"]:
                info = connection.ecs.list_tags_for_resource(resourceArn=arn)
                tags = {i["key"]: i["value"] for i in info["tags"]}
                if tags.get("createdBy") == "ecs-deplojo":
                    yield arn

    logger.info("Deregistering old task definitions")
    for service_name, task_definition in task_definitions.items():
        logger.info(" - %s", task_definition.family)

        num = 0

        for arn in yield_arns(task_definition.family):
            num += 1
            if arn != task_definition.arn:
                connection.ecs.deregister_task_definition(taskDefinition=arn)

            if num > 10:
                break



import copy
import json
import operator
import os.path
import typing
from string import Template


class TaskDefinition:
    """A TaskDefinition exists out of a set of containers."""

    def __init__(self, data):
        self._data = data

    @classmethod
    def load(cls, fh) -> "TaskDefinition":
        data = json.load(fh)
        return cls(data)

    def as_dict(self):
        """Output the TaskDefinition in a boto3 compatible format.

        See the boto3 documentation on `ECS.Client.register_task_definition`.
        """
        result = copy.deepcopy(self._data)
        for container in result["containerDefinitions"]:
            container["environment"] = sorted(
                [
                    {"name": k, "value": str(v)}
                    for k, v in container.get("environment", {}).items()
                ],
                key=operator.itemgetter("name"),
            )
            container["secrets"] = sorted(
                [
                    {"name": k, "valueFrom": str(v)}
                    for k, v in container.get("secrets", {}).items()
                ],
                key=operator.itemgetter("name"),
            )
        return result

    def apply_variables(self, variables: typing.Dict[str, str]):
        """Interpolate all the variables used in the task definition"""
        for container in self.container_definitions:
            container["image"] = Template(container["image"]).substitute(variables)

    def apply_overrides(self, overrides):
        """Apply overrides for all containers within this task definition."""
        for container in self.container_definitions:
            container_overrides = overrides.get(container["name"], {})
            for key, value in container_overrides.items():
                if key in container and isinstance(container[key], list):
                    container[key].extend(value)
                elif key in container and isinstance(container[key], dict):
                    container[key].update(value)
                else:
                    container[key] = value

    def set_environment(self, env: typing.Dict[str, str]):
        """Interpolate all the variables used in the task definition"""
        for container in self.container_definitions:
            container["environment"] = env

    def set_secrets(self, secrets: typing.Dict[str, str]):
        """Interpolate all the secrets used in the task definition.

        Secrets will be fetched from the AWS Parameter store and injected in the
        environment variables during container startup

        """
        for container in self.container_definitions:
            container["secrets"] = secrets

    def __str__(self):
        if self._data.get("name"):
            return self._data.get("name")
        return "(unregistered)"

    def __eq__(self, other: object):
        if not isinstance(other, TaskDefinition):
            return False
        return self._data == other._data

    def __repr__(self):
        return json.dumps(self._data)

    @property
    def tags(self) -> typing.List[typing.Dict[str, str]]:
        return self._data.get("tags")

    @tags.setter
    def tags(self, value: typing.List[typing.Dict[str, str]]):
        self._data["tags"] = value

    @property
    def family(self) -> str:
        return self._data.get("family")

    @family.setter
    def family(self, value: str):
        self._data["family"] = value

    @property
    def revision(self) -> int:
        return self._data.get("revision")

    @revision.setter
    def revision(self, value: int):
        self._data["revision"] = value

    @property
    def name(self) -> str:
        return self._data.get("name")

    @name.setter
    def name(self, value: str):
        self._data["name"] = value

    @property
    def task_role_arn(self) -> str:
        return self._data.get("taskRoleArn")

    @task_role_arn.setter
    def task_role_arn(self, value: str):
        self._data["taskRoleArn"] = value

    @property
    def execution_role_arn(self) -> str:
        return self._data.get("executionRoleArn")

    @execution_role_arn.setter
    def execution_role_arn(self, value: str):
        self._data["executionRoleArn"] = value

    @property
    def arn(self) -> str:
        return self._data.get("arn")

    @arn.setter
    def arn(self, value: str):
        self._data["arn"] = value

    @property
    def container_definitions(self):
        return self._data.get("containerDefinitions")

    @container_definitions.setter
    def container_definitions(self, value):
        self._data["containerDefinitions"] = value

    @property
    def network_mode(self) -> typing.Optional[str]:
        return self._data.get("networkMode", None)

    @network_mode.setter
    def network_mode(self, value):
        self._data["networkMode"] = value


def generate_task_definitions(
    config, template_vars, base_path, output_path=None
) -> typing.Dict[str, TaskDefinition]:
    """Generate the task definitions

    :parameter config: The yaml config contents
    :parameter template_vars: Key-Value dict with template replacements
    :parameter base_path: The base path (location of the config file)
    :parameter output_path: Optional path to write the task definitions to.
    :rtype dict:

    """
    task_definitions = {}

    for name, info in config["task_definitions"].items():
        # Create a copy of the environment dict so that it can safely be
        # modified.
        env_vars = copy.deepcopy(config.get("environment", {}))

        # Environment groups
        env_group = info.get("environment_group")
        if env_group:
            env_vars.update(config["environment_groups"][env_group])

        overrides = info.get("overrides", {})
        definition = generate_task_definition(
            filename=info["template"],
            environment=env_vars,
            template_vars=template_vars,
            overrides=overrides,
            name=name,
            base_path=base_path,
            task_role_arn=info.get("task_role_arn"),
            secrets=config.get("secrets", {}),
            execution_role_arn=info.get("execution_role_arn"),
        )
        if output_path:
            write_task_definition(name, definition, output_path)
        task_definitions[name] = definition

    return task_definitions


def generate_task_definition(
    filename: str,
    environment: typing.Dict[str, str],
    template_vars,
    overrides,
    name,
    base_path=None,
    task_role_arn=None,
    secrets: typing.Dict[str, str] = {},
    execution_role_arn=None,
) -> TaskDefinition:
    """Generate the task definitions."""
    if base_path:
        filename = os.path.join(base_path, filename)

    with open(filename, "r") as fh:
        task_definition = TaskDefinition.load(fh)

    task_definition.family = name
    if task_role_arn:
        task_definition.task_role_arn = task_role_arn

    if execution_role_arn:
        task_definition.execution_role_arn = execution_role_arn

    # If no hostname is specified for the container we set it ourselves to
    # `{family}-{container-name}-{num}`
    # Skip this when network_mode == awsvpc, not supported by AWS.
    if task_definition.network_mode not in ["awsvpc"]:
        num_containers = len(task_definition.container_definitions)
        for container in task_definition.container_definitions:
            hostname = task_definition.family
            if num_containers > 1:
                hostname += "-%s" % container["name"].replace("_", "-")
            container.setdefault("hostname", hostname)

    task_definition.set_environment(environment)
    if secrets:
        task_definition.set_secrets(secrets)
    task_definition.apply_variables(template_vars)
    task_definition.apply_overrides(overrides)
    task_definition.tags = [{"key": "createdBy", "value": "ecs-deplojo"}]

    return task_definition


def write_task_definition(name: str, definition: TaskDefinition, output_path) -> None:
    filename = os.path.join(output_path, "%s.json" % name)
    with open(filename, "w") as fh:
        json.dump(definition.as_dict(), fh, indent=4)
